# Copyright (c) Contributors 2020
# See License for details
"""
Provides seashore version information.
"""

# This file is auto-generated! Do not edit!
# Use `python -m incremental.update seashore` to change this file.

from incremental import Version

__version__ = Version('seashore', 20, 6, 0)
__all__ = ["__version__"]
